#![no_main]
#![no_std]

extern crate cortex_m;
extern crate cortex_m_rt as rt;
//extern crate panic_reset;
extern crate panic_semihosting;
extern crate stm32f4xx_hal as hal;

mod packet;

#[rtic::app(device = stm32f4xx_hal::pac, peripherals = true)]
mod app {
    use hal::dma;
    use hal::gpio::{self, Speed};
    use hal::pac::{self, SPI1, TIM5};
    use hal::prelude::*;
    use hal::spi;
    use hal::spi::{Mode, Phase, Polarity, Spi};
    use hal::timer::Delay;
    use rf4463::config::RADIO_CONFIG_500_2;
    use rf4463::Rf4463;
    use ringbuffer::{ConstGenericRingBuffer, RingBuffer, RingBufferRead, RingBufferWrite};

    use crate::packet::Packet;
    use crate::packet::PACKET_LEN;

    // in # packets
    const BUFFER_LEN: usize = 50;

    // in bytes
    // explicitly make this smaller than a packet so that we don't buffer things in the tx DMA for a while
    const PI_RX_BUFFER_LEN: usize = 256;

    // in bytes
    const PI_TX_BUFFER_LEN: usize = 1;

    const MODE: Mode = Mode {
        polarity: Polarity::IdleLow,
        phase: Phase::CaptureOnFirstTransition,
    };

    type Radio = Rf4463<
        PACKET_LEN,
        Spi<SPI1>,
        gpio::Pin<'B', 1, gpio::Output>,
        gpio::Pin<'B', 0, gpio::Output>,
        Delay<TIM5, 1000000>,
    >;

    type RxTransfer = dma::Transfer<
        dma::StreamX<pac::DMA1, 3>,
        0,
        spi::Rx<pac::SPI2>,
        dma::PeripheralToMemory,
        &'static mut [u8; PI_RX_BUFFER_LEN],
    >;

    type TxTransfer = dma::Transfer<
        dma::StreamX<pac::DMA1, 4>,
        0,
        spi::Tx<pac::SPI2>,
        dma::MemoryToPeripheral,
        &'static mut [u8; PI_TX_BUFFER_LEN],
    >;

    #[derive(Debug)]
    enum SlaveCmd {
        SendPacket,
        GetTemp,
    }

    impl SlaveCmd {
        pub fn from_u8(i: u8) -> Option<Self> {
            match i {
                0x01 => Some(Self::SendPacket),
                0x02 => Some(Self::GetTemp),
                _ => None,
            }
        }
    }

    #[allow(clippy::large_enum_variant)]
    pub enum SlaveState {
        Idle,
        RecvPacket(Packet),
    }

    #[shared]
    struct Shared {
        radio: Radio,
        radio_temp: f32,
        tx_buf: ConstGenericRingBuffer<Packet, BUFFER_LEN>,

        pi_tx: TxTransfer,
        other_tx_buf: Option<&'static mut [u8; PI_TX_BUFFER_LEN]>,
    }

    #[local]
    struct Local {
        radio_irq: gpio::Pin<'B', 2, gpio::Input>,
        state: SlaveState,

        pi_rx: RxTransfer,
        other_rx_buf: Option<&'static mut [u8; PI_RX_BUFFER_LEN]>,
    }

    #[init]
    fn init(mut ctx: init::Context) -> (Shared, Local, init::Monotonics) {
        let rcc = ctx.device.RCC.constrain();

        let gpioa = ctx.device.GPIOA.split();
        let gpiob = ctx.device.GPIOB.split();

        // setup clocks
        let clocks = rcc
            .cfgr
            .use_hse(25.MHz())
            .sysclk(100.MHz())
            .pclk1(48.MHz())
            .pclk2(48.MHz())
            .freeze();
        let mut sys_cfg = ctx.device.SYSCFG.constrain();

        // setup 4463 spi

        let mosi = gpioa.pa7.into_alternate().speed(Speed::VeryHigh);
        let miso = gpioa.pa6.into_alternate().speed(Speed::VeryHigh);
        let sclk = gpioa.pa5.into_alternate().speed(Speed::VeryHigh);

        let sdn = gpiob.pb1.into_push_pull_output();
        let cs = gpiob.pb0.into_push_pull_output();
        let mut radio_irq = gpiob.pb2.into_pull_up_input();

        let spi = Spi::new(ctx.device.SPI1, (sclk, miso, mosi), MODE, 10.MHz(), &clocks);

        let delay = ctx.device.TIM5.delay_us(&clocks);
        let radio = Rf4463::new(spi, sdn, cs, delay, &mut RADIO_CONFIG_500_2.clone()).unwrap();

        radio_irq.make_interrupt_source(&mut sys_cfg);
        radio_irq.enable_interrupt(&mut ctx.device.EXTI);
        radio_irq.trigger_on_edge(&mut ctx.device.EXTI, gpio::Edge::Falling);

        // setup Pi SPI
        let pi_mosi = gpiob.pb15.internal_resistor(gpio::Pull::Up);
        let pi_miso = gpiob.pb14.internal_resistor(gpio::Pull::Down);
        let pi_sclk = gpiob.pb13.internal_resistor(gpio::Pull::Down);

        // TODO nss?

        let mut pi_spi = ctx
            .device
            .SPI2
            .spi_slave((pi_sclk, pi_miso, pi_mosi, None), MODE);
        pi_spi.set_internal_nss(false);

        let (tx, rx) = pi_spi.use_dma().txrx();

        let streams = dma::StreamsTuple::new(ctx.device.DMA1);

        let tx_buffer =
            cortex_m::singleton!(: [u8; PI_TX_BUFFER_LEN] = [0x02; PI_TX_BUFFER_LEN]).unwrap();
        let other_tx_buf =
            Some(cortex_m::singleton!(: [u8; PI_TX_BUFFER_LEN] = [0u8; PI_TX_BUFFER_LEN]).unwrap());

        let mut pi_tx = dma::Transfer::init_memory_to_peripheral(
            streams.4,
            tx,
            tx_buffer,
            None,
            dma::config::DmaConfig::default()
                .memory_increment(false)
                .fifo_enable(true)
                .fifo_error_interrupt(true),
        );

        pi_tx.start(|_tx| {});

        let rx_buffer =
            cortex_m::singleton!(: [u8; PI_RX_BUFFER_LEN] = [0u8; PI_RX_BUFFER_LEN]).unwrap();
        let other_rx_buf =
            Some(cortex_m::singleton!(: [u8; PI_RX_BUFFER_LEN] = [0u8; PI_RX_BUFFER_LEN]).unwrap());

        let mut pi_rx = dma::Transfer::init_peripheral_to_memory(
            streams.3,
            rx,
            rx_buffer,
            None,
            dma::config::DmaConfig::default()
                .memory_increment(true)
                .fifo_enable(true)
                .fifo_error_interrupt(true)
                .transfer_complete_interrupt(true),
        );

        pi_rx.start(|_rx| {});

        (
            Shared {
                radio,
                radio_temp: 0.0,
                tx_buf: ConstGenericRingBuffer::new(),

                pi_tx,
                other_tx_buf,
            },
            Local {
                radio_irq,
                state: SlaveState::Idle,

                pi_rx,
                other_rx_buf,
            },
            init::Monotonics(),
        )
    }

    #[idle(shared=[radio, tx_buf, radio_temp, pi_tx, other_tx_buf])]
    fn idle(mut ctx: idle::Context) -> ! {
        let mut i = 0;
        let mut iterations_since_last_packet = 0;
        loop {
            if let Some(pkt) = ctx.shared.tx_buf.lock(|buf| buf.dequeue()) {
                if iterations_since_last_packet >= 1000 {
                    // if we haven't transmitted in a while, the receiver's AGC is going to be messed up
                    // send 5 packets just so it can calibrate before we actually send it
                    for _ in 0..5 {
                        const GARBAGE: [u8; PACKET_LEN] = [0x00; PACKET_LEN];
                        ctx.shared.radio.lock(|r| r.start_tx(GARBAGE).unwrap());

                        // wait for radio to be idle again
                        loop {
                            if ctx.shared.radio.lock(|r| r.is_idle()) {
                                break;
                            }
                        }
                    }
                }

                ctx.shared.radio.lock(|r| r.start_tx(pkt.data).unwrap());

                // wait for radio to be idle again
                loop {
                    if ctx.shared.radio.lock(|r| r.is_idle()) {
                        break;
                    }
                }

                // update our SPI bus to say if we have free space or not
                let b = if ctx.shared.tx_buf.lock(|b| b.len() >= b.capacity() - 2) {
                    0x01 // not enough space for packet
                } else {
                    0x02 // space for packet
                };

                set_spi_tx_byte(b, &mut ctx.shared.pi_tx, &mut ctx.shared.other_tx_buf);

                iterations_since_last_packet = 0;
            } else {
                iterations_since_last_packet += 1;
            }

            i += 1;

            // get temp every 500 packets, or whenever when idle
            if i >= 500 {
                i = 0;

                if let Ok(new_temp) = ctx.shared.radio.lock(|r| r.get_temp()) {
                    ctx.shared.radio_temp.lock(|cur_temp| *cur_temp = new_temp);
                }
            }
        }
    }

    #[task(binds = DMA1_STREAM3, priority = 2, local = [state, pi_rx, other_rx_buf], shared = [tx_buf, pi_tx, other_tx_buf])]
    fn pi_spi_recv(mut ctx: pi_spi_recv::Context) {
        let rx = ctx.local.pi_rx;

        let (old_buf, _) = rx
            .next_transfer(ctx.local.other_rx_buf.take().unwrap())
            .unwrap();

        for b in old_buf.iter() {
            handle_incoming_byte(
                *b,
                ctx.local.state,
                &mut ctx.shared.tx_buf,
                &mut ctx.shared.pi_tx,
                &mut ctx.shared.other_tx_buf,
            );
        }

        *ctx.local.other_rx_buf = Some(old_buf);

        rx.clear_transfer_complete_interrupt();
        rx.clear_fifo_error_interrupt();
    }

    #[task(binds = DMA1_STREAM4, priority = 2, shared = [pi_tx])]
    fn pi_spi_tx(mut ctx: pi_spi_tx::Context) {
        ctx.shared.pi_tx.lock(|tx| tx.clear_fifo_error_interrupt());
    }

    #[task(binds = EXTI2, shared = [radio], local = [radio_irq])]
    fn radio_irq(mut ctx: radio_irq::Context) {
        ctx.shared.radio.lock(|r| r.interrupt()).unwrap();

        let irq = ctx.local.radio_irq;
        if irq.is_high() {
            irq.clear_interrupt_pending_bit();
        }
    }

    fn handle_incoming_byte<
        T: rtic::Mutex<T = ConstGenericRingBuffer<Packet, BUFFER_LEN>>,
        A: rtic::Mutex<T = TxTransfer>,
        B: rtic::Mutex<T = Option<&'static mut [u8; PI_TX_BUFFER_LEN]>>,
    >(
        x: u8,
        state: &mut SlaveState,
        buf: &mut T,
        tx: &mut A,
        other_tx: &mut B,
    ) {
        match state {
            SlaveState::Idle => {
                if let Some(cmd) = SlaveCmd::from_u8(x) {
                    match cmd {
                        SlaveCmd::SendPacket => {
                            let b = if buf.lock(|b| b.len() >= b.capacity() - 2) {
                                0x01 // not enough space for packet
                            } else {
                                0x02 // space for packet
                            };

                            set_spi_tx_byte(b, tx, other_tx);

                            *state = SlaveState::RecvPacket(Packet::default());
                        }

                        SlaveCmd::GetTemp => {
                            todo!();

                            /*
                                        let temp = ctx.shared.radio_temp.lock(|x| *x);
                                        usart.bwrite_all(&temp.to_le_bytes()).unwrap();
                            */
                        }
                    }
                }
            }

            SlaveState::RecvPacket(pkt) => {
                pkt.push(x);
                if pkt.is_complete() {
                    buf.lock(|b| b.enqueue(pkt.clone()));

                    *state = SlaveState::Idle;
                }
            }
        };
    }

    fn set_spi_tx_byte<
        A: rtic::Mutex<T = TxTransfer>,
        B: rtic::Mutex<T = Option<&'static mut [u8; PI_TX_BUFFER_LEN]>>,
    >(
        b: u8,
        tx: &mut A,
        other_tx: &mut B,
    ) {
        other_tx.lock(|t| t.as_mut().unwrap()[0] = b);

        other_tx.lock(|t| {
            let (tx, _) =
                tx.lock(|tx_transfer| tx_transfer.next_transfer(t.take().unwrap()).unwrap());

            t.replace(tx);
        });
    }
}

// TODO need an interrupt to clear TX errors
